#
# Copyright (c) 1999 The University of Paris VI
# All rights reserved.
# 
# Permission to use and copy this software in source and binary forms
# is hereby granted, provided that the above copyright notice, this
# paragraph and the following disclaimer are retained in any copies
# of any part of this software and that the University of Paris VI is
# acknowledged in all documentation pertaining to any such copy
# or derivative work. The name of the University of Paris VI may not
# be used to endorse or promote products derived from this software
# without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
# THE UNIVERSITY OF PARIS VI BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#
# Contributed by the RP Research Group, http://www.lip6.fr/rp
#

# Work with 2 classes ONLY

# Load main files for script 
#
source in_process.tcl
source post_process.tcl
source wred_lib.tcl
source core_lib.tcl
source ns-airs.tcl

#====================================================
# Simulation Parameters
#====================================================
# Set options table
set opt(trace)	"tr" ; # suffix for trace file 
set opt(monitor) "mt" ; # suffix for monitor file
set opt(interval) 5
set opt(graphic)  gnuplot
set opt(random-seed) 1
set opt(bottle-queue-length) bw-delay-product
set opt(bottle-delay)  50ms
set opt(class-delay) 3ms

#====================================================
# TestSuite Base Class
#====================================================

Class TestSuite 

TestSuite instproc init {} {
	$self instvar ns_  test_ stoptime_ post_ rng_ avg_len_
	$self instvar b1 b2
	$self instvar cs_count_ ;# transfert count
	$self instvar client_count_ ;# number of created clients
	$self instvar client_onduty_
	$self instvar bottleneck_link_ seed_
	$self instvar nametracefile_ namegraphfile_ namemonitorfile_
	$self instvar tracefilefd_
	global opt

	set ns_ [new Simulator]

        # Seed random no. generator; ns-random with arg of :
	# 0 heuristically chooses a random number that changes on each invocation.
	# 1 deterministically, same sequence of random number on each invocation
	set rng_ [new RNG]
	$rng_ seed $opt(random-seed)
	$rng_ next-random
	set seed_  $opt(random-seed)

	# Compute optimal (?) bottleneck queue size
	# as the bw-delay product.
	if {$opt(bottle-queue-length) == "bw-delay-product"} {
		set opt(bottle-queue-length) \
                   [expr ([$ns_ bw_parse $opt(bottle-bw)] \
			 * ([$ns_ delay_parse $opt(bottle-delay)] + 2*[$ns_ delay_parse $opt(class-delay)]) \
			 + $opt(pkt-size)*8 - 1)\
		    / ($opt(pkt-size) * 8)]
		puts "optimal bw queue size: [expr round($opt(bottle-queue-length))]"
	}
	
	Queue set limit_ [expr round($opt(bottle-queue-length))]
	
	# expanded addresses (21 bits nodes).
	$ns_ rtproto Manual
#	$ns_ set-address-format expanded
	$ns_ set-address 15 8
	$ns_ expand-port-field-bits 11


	set cs_count_ 0 ;	# transfert count
	set client_count_ 0;    # number of created clients

	# set others parameters
	set name [$self info class]
        set lname [split $name /]
        set test_ [lindex $lname 1]
	set stoptime_ $opt(stoptime)
	set avg_len_ $opt(L)

	# set Traffic parameters
	Agent/TCP set window_ 1000
	Agent/TCP set tcpTick_ 0.1
	
	if {$opt(pkt-size) > 0}  {
		Agent/TCP set packetSize_ $opt(pkt-size)
	}
        Agent/TCP set ecn_ 0	

	# Set names for differents useful files
	set nametracefile_ $test_.$opt(trace)
	set namemonitorfile_ $test_.$opt(monitor)
	set namegraphfile_ $test_
	

	# Open file for transfert trace
       	if [file exists $nametracefile_] {
        	catch "exec rm -f $nametracefile_"
        }
        
        set tracefilefd_ [open $nametracefile_ w]

        
	# create topology
	$self create_topo
     	set bottleneck_link_ [$ns_ link $b1 $b2]

	# create forward traffic
	$self schedule_initial_traffic $opt(lambda1) $opt(lambda2)

	# add traffic monitor
#	$self config_flow_monitor $opt(interval)

	# create a post process environment
	set post_ [new PostProcess/$opt(graphic) $test_ $nametracefile_ $namegraphfile_]

}


TestSuite instproc schedule_initial_traffic {lambda1 lambda2} {
	$self instvar ns_ s1 s2 r1 r2
	$self instvar idle_clients_ ;# array of lists for idle client on each class 
	$self instvar rng_ client_onduty_
	$self instvar finished_

	for {set i 1} {$i <= 2} {incr i} {
		set idle_clients_($i) ""
 		set client_onduty_($i) 0;   # number of active clients
 		set finished_($i) 0;   # number of finished transfert
            	
		eval "set lambda \$lambda$i"
		if {$lambda != 0.} {
		   #flows start randomly between 0 and 10 secs
		   set startt [$rng_ uniform 0 10]
		   $ns_ at $startt "$self schedule_continuing_traffic $i $lambda 0"
		}
	}
}


TestSuite instproc schedule_continuing_traffic {traffic_class lambda delay} {
	$self instvar ns_ rng_ 

	$self start_a_client $traffic_class $delay
	# schedule the next one
#	set next [expr [$ns_ now]+([$rng_ exponential]/$lambda)]; #??? mysterious
#	set inter_delay [expr 1/[$rng_ exponential $lambda]]

	set inter_delay [$rng_ exponential [expr 1/$lambda]]

	set next [expr [$ns_ now]+$inter_delay]
	$ns_ at $next "$self schedule_continuing_traffic $traffic_class $lambda $inter_delay"
}



#
# create a tcp source/dst with a ftp traffic generator
#
TestSuite instproc create_a_client {traffic_class  node_s node_r} {
	$self instvar node_  sources_ app_ b1
	$self instvar ns_ client_count_ 
	global opt
	
	set now [$ns_ now]
	
	set newnode  [$ns_ node]
	set client_id [incr client_count_]
	set node_($client_id) $newnode
	
	puts "t: [format %.3f $now ] create_a_client $client_id"

	$ns_ duplex-link $newnode $node_s \
		$opt(M) $opt(class-delay) DropTail
		
	#Add routing
	 $newnode add-route-to-adj-node -default $node_s
	 $node_s    add-route-to-adj-node $newnode
	 set link [$ns_ link $b1 $node_s]
	 $b1 add-route  [$newnode id] [$link head]

	
	# create sources and sinks
	set clients [$ns_ create-connection-list TCP/Reno $newnode TCPSink $node_r 0]
	set sources_($client_id) [lindex $clients 0]
	set dst_agent [lindex $clients 1]
	
	$sources_($client_id) set dst_agent_ $dst_agent
        
        set prio  [expr $traffic_class*pow(16,6)]
        $sources_($client_id) set prio_ $prio
	$dst_agent set prio_ $prio
	
	# Set up a callback when this client ends.
	$sources_($client_id) proc done {} "$self finish_a_client $traffic_class $client_id"

        set app_($client_id) [new Application/FTP]
        $app_($client_id) attach-agent $sources_($client_id)
        

        return $client_id
}



TestSuite instproc start_a_client {traffic_class inter_delay} {
	$self instvar idle_clients_  sources_  app_ tracefilefd_
	$self instvar ns_ rng_ avg_len_
	$self instvar cs_count_ client_onduty_
	$self instvar s1 s2 r1 r2


	set now [$ns_ now]
	
	# Have we still available client?
	set x [llength $idle_clients_($traffic_class)]	
	if {$x == 0} {
		set node_s [eval "\$self set s$traffic_class"] 
		set node_r [eval "\$self set r$traffic_class"] 

		lappend idle_clients_($traffic_class) \
			[$self create_a_client $traffic_class $node_s $node_r]
	} else {
#		puts "t: [format %.3f $now ] re_use client [lindex $idle_clients_($traffic_class) 0]"
	}
	
	set i [lindex $idle_clients_($traffic_class) 0]
	set idle_clients_($traffic_class) [lrange $idle_clients_($traffic_class) 1 end]

	# Reset the connection.
	$sources_($i) reset
	[$sources_($i) set dst_agent_] reset
	
	incr cs_count_
	$sources_($i) set transfert_count_    $cs_count_
	# to use flow monitor
	$sources_($i) set fid_ $traffic_class
	$sources_($i) set startt_ $now
	$sources_($i) set inter_delay_ $inter_delay
	
	# size of the file to transfert
	$sources_($i) set transfert_size_ [expr round ([$rng_ exponential $avg_len_])]
	set len [$sources_($i) set transfert_size_]

	# Start traffic for that client.
	$app_($i) send $len

#	$sources_($i) advanceby $len
	set now [$ns_ now]
	
	# Write a trace
	###############
	puts $tracefilefd_ "t: [format %.3f $now ] starting_client $i transfert_id $cs_count_ class $traffic_class $len bytes inter_delay [format %.3f $inter_delay] [Queue set limit_ ]"

	incr client_onduty_($traffic_class)
 	
}


TestSuite instproc finish_a_client {traffic_class i} {
	$self instvar ns_ client_onduty_ 
	$self instvar idle_clients_ sources_
	$self instvar tracefilefd_ finished_

	incr client_onduty_($traffic_class) -1

	set now [$ns_ now]
	set delta [expr $now - [$sources_($i) set startt_]]

	# Write a trace
	###############
	puts $tracefilefd_ "t: [format %.3f $now] finished_client $i transfert_id [$sources_($i) set transfert_count_] class $traffic_class  [$sources_($i) set transfert_size_] bytes inter_delay [format %.3f [$sources_($i) set inter_delay_]] duration [format %.3f $delta] When_active_client $client_onduty_($traffic_class)"

	lappend idle_clients_($traffic_class) $i
	
	incr finished_($traffic_class)
	$self test-finish
}




TestSuite instproc create_topo {} {
	$self instvar ns_ s1 s2 r1 r2 b1 b2
	$self instvar test_ 
	global opt
#
# S1	      R1	
#   \	     /	 
#    \	    /
#    B1---B2
#    /	    \	
#   /	     \
# S2	      R2	
#
       	set b1 [$ns_ node]
    	set b2 [$ns_ node]
	set s1 [$ns_ node]
    	set s2 [$ns_ node]
    	set r1 [$ns_ node]
    	set r2 [$ns_ node]
   
    
    set bw_class [expr $opt(max-client-count)*[$ns_ bw_parse $opt(M)]]
    $ns_ duplex-link $s2 $b1 $bw_class $opt(class-delay) DropTail
    $ns_ duplex-link $s1 $b1 $bw_class $opt(class-delay) DropTail	
    $ns_ duplex-link $b1 $b2 $opt(bottle-bw) $opt(bottle-delay) $test_
    $ns_ duplex-link $b2 $r1 $bw_class $opt(class-delay) DropTail
    $ns_ duplex-link $b2 $r2 $bw_class $opt(class-delay) DropTail	

    	# Bottlenecks need large routing tables.
	[$b1 set classifier_] resize 511
	[$s2 set classifier_] resize 255
	[$s1 set classifier_] resize 255

	# Default routes to the other.
	$b1 add-route-to-adj-node -default $b2
	$b2 add-route-to-adj-node -default $b1
	
	# Add routing in all directions
	$r1 add-route-to-adj-node -default $b2
	$r2 add-route-to-adj-node -default $b2
	$s1 add-route-to-adj-node -default $b1
	$s2 add-route-to-adj-node -default $b1
	
	$b2 add-route-to-adj-node $r1
	$b2 add-route-to-adj-node $r2
	
	$b1 add-route-to-adj-node $s1
	$b1 add-route-to-adj-node $s2

}


TestSuite instproc config_flow_monitor {interval } {
        $self instvar ns_  bottleneck_link_
        $self instvar monitor_
	$self instvar namemonitorfile_ 
	global opt

        set monitor_ [new MonitorFlowPrio $ns_ $bottleneck_link_ $opt(bottle-bw)]
        $monitor_ monitor-link 2
        
        # clear file for monitor trace
       	if [file exists $namemonitorfile_] {
        	catch "exec rm -f $namemonitorfile_"
        }
        $monitor_ trace_bdep  $interval $namemonitorfile_
}


TestSuite instproc run {} {
	$self instvar ns_  stoptime_ seed_

#	$ns_ at $stoptime_ "$self finish" 

    	ns-random seed_; # 0 for heuristic seed source
    	$ns_ run
}


TestSuite instproc test-finish {} {
	$self instvar ns_  stoptime_ finished_

	if { $finished_(1) >= $stoptime_ && $finished_(2) >= $stoptime_ } {
		
		set now [$ns_ now]
		$ns_ at $now "$self finish" 
		return
	}
	if { $finished_(1) >= (5*$stoptime_) || $finished_(2) >= (5*$stoptime_) } {
		
		set now [$ns_ now]
		$ns_ at $now "$self finish" 
	}
}

TestSuite instproc finish {} {
	$self instvar ns_ post_ namegraphfile_ nametracefile_
	$self instvar tracefilefd_ monitor_
 	global opt 

       $ns_ flush-trace
       
       # close tracefile
       close $tracefilefd_
        
        if {[info exists post_]} {
		$post_ plot_bytes $opt(bottle-bw)
        }
        
              
        if {[info exists monitor_]} {
                set tmp tmp[pid]
        	catch "exec rm -f [glob tmp*]"

		# open graph files        
        	set sumTPfd [open $opt(summary-file).tmp a]
		$monitor_ finish $tmp $namegraphfile_.bneck $sumTPfd
	        close $sumTPfd
  	        catch "exec rm -f $tmp"
        }

      	# Erase  working files
       	if [file exists $nametracefile_] {
#	       	catch "exec rm -f $nametracefile_"
        }

      	if [file exists $namegraphfile_] {
#	    	catch "exec rm -f $namegraphfile_"
        }
 

        exit 0
}

#====================================================
# Specific test Class
#====================================================

Class Test/WFQ -superclass TestSuite

Test/WFQ instproc create_topo {} {
	$self instvar ns_
	$self instvar b1 b2
	global opt
		
	$self next
	
	# for traffic class 1
	$ns_ install-phb $b1 $b2 1 Queue/DropTail $opt(phi); 
 	# for traffic class 2
 	$ns_ install-phb $b1 $b2 2 Queue/DropTail 1
}


Class Test/WRED -superclass TestSuite

Test/WRED instproc create_topo {} {
    $self instvar ns_
    $self instvar b1 b2
    global opt	
    $self next

    set bottl_1 [$ns_ link $b1 $b2]
    set bottl_q1 [$bottl_1 queue]

    $bottl_q1 set ECN 0
    $bottl_q1 set thresh_l 5
    $bottl_q1 set maxthresh_l 40
    $bottl_q1 set linterm_l 1
    $bottl_q1 set thresh_h 50
    $bottl_q1 set maxthresh_h 250
    $bottl_q1 set linterm_h 50
    $bottl_q1 set limit_ 300
    $bottl_q1 set mode 0
    $bottl_q1 set byte_mode false
    $bottl_q1 set q_weight 0.002
    $bottl_q1 set alpha_ 1
    $bottl_q1 set beta_ 1 
    $bottl_q1 set trace_int_ 200
# queue tracind is disabled because jgraph is not installed
#    enable_tracequeue_WRED $bottl_q1  
}
#
# create a tcp source/dst with a ftp traffic generator
#
Test/WRED instproc create_a_client {traffic_class  node_s node_r} {
	$self instvar node_  sources_ app_ b1
	$self instvar ns_ client_count_ tracefilefd_
	global opt
	
	set now [$ns_ now]
	
	set newnode  [$ns_ node]
	set client_id [incr client_count_]
	set node_($client_id) $newnode
	
	puts "t: [format %.3f $now ] create_a_client $client_id"

	$ns_ duplex-link $newnode $node_s \
		$opt(M) $opt(class-delay) DropTail
		
	#Add routing
	 $newnode add-route-to-adj-node -default $node_s
	 $node_s    add-route-to-adj-node $newnode
	 set link [$ns_ link $b1 $node_s]
	 $b1 add-route  [$newnode id] [$link head]

	
	# create sources and sinks
	set clients [$ns_ create-connection-list TCP/Reno $newnode TCPSink $node_r 0]
	set sources_($client_id) [lindex $clients 0]
	set dst_agent [lindex $clients 1]
	
	$sources_($client_id) set dst_agent_ $dst_agent

        if {$traffic_class == 2} { 
           set prio [expr pow(16,6)]
           #puts $tracefilefd_ "prio $prio"
	} else {
	    set prio 0
	}
        $sources_($client_id) set prio_ $prio
	$dst_agent set prio_ $prio
	
	# Set up a callback when this client ends.
	$sources_($client_id) proc done {} "$self finish_a_client $traffic_class $client_id"

        set app_($client_id) [new Application/FTP]
        $app_($client_id) attach-agent $sources_($client_id)
        

        return $client_id
}


Test/WRED instproc finish {} {
	$self instvar ns_ post_ namegraphfile_ nametracefile_
	$self instvar tracefilefd_
        $self instvar b1 b2 stoptime_ opt

 	global opt 

       $ns_ flush-trace
       
       # close tracefile
       close $tracefilefd_
        
        if {[info exists post_]} {
		$post_ plot_bytes $opt(bottle-bw)
        }
        
      	# Erase  working files
       	if [file exists $nametracefile_] {
#	       	catch "exec rm -f $nametracefile_"
        }

      	if [file exists $namegraphfile_] {
#	    	catch "exec rm -f $namegraphfile_"
        }

# queue tracing is disabled because jgraph is not installed
    	#set now [$ns_ now]
        #set bottl_1 [$ns_ link $b1 $b2]
        #set bottl_q1 [$bottl_1 queue]


        #plotQueue_WRED $namegraphfile_ $bottl_q1 $now

	#set suffix [lindex [split $opt(rho1) .] 1]
	#set fnam "queue"
	#set suffix2 ".ps"
	#set qfile "queue.ps"
	#catch "exec mv $qfile $fnam$suffix$suffix2"
        exit 0
}



Class Test/DropTail -superclass TestSuite


#====================================================
# Procedures Definition
#====================================================


# Procedure Usage ()
proc usage {} {
	global argv0
	puts stderr "usage: ns $argv0 phi on1 off1 on2 off2 bw M N L rho1 rho2 file <test>"
	exit 1
}


proc runTest {test} {
	switch $test {
		WFQ -
	        WRED -
		DropTail - 
		PQCM {
			set t [new Test/$test]
		}
		default {
			puts stderr "Unkown test $test"
			exit 1
		}
	}
	$t run
}


#=======================================
##                 MAIN               ##
#=======================================

	global argc argv
	
	if {$argc < 12} {
		usage
	}
	
	set opt(phi) [lindex $argv 0]
	set opt(lambda1) [lindex $argv 1]
	set opt(lambda2) [lindex $argv 2]
	set opt(max-client-count) [lindex $argv 3]
	set opt(pkt-size) [lindex $argv 4]
	set opt(bottle-bw) [lindex $argv 5]
	set opt(M) [lindex $argv 6]
	set opt(begin-compute) [lindex $argv 7]
	set opt(L) [lindex $argv 8]
	set opt(rho1) [lindex $argv 9]
	set opt(rho2) [lindex $argv 10]
	set opt(summary-file)  [lindex $argv 11]
	set opt(stoptime)  [lindex $argv 12]
	set test      [lindex $argv 13]

	runTest $test
